/********************************************************************************
*                                                                               *
*                      F X   M O V I E   M A N A G E R                          *
*                                                                               *
*********************************************************************************
* Copyright(C) 2008-2012 by Luigi Ziviani <ziviani_luigi@yahoo.it>.             *
* All Rights Reserved.                                 v 5.6                    *
*********************************************************************************
* Id: fxplayer.c,                                      2008/01/01               *
*********************************************************************************
* This file is part of FXMOVIEMANAGER                                           *
*                                                                               *
* FXMOVIEMANAGER is free software; you can redistribute it and/or modify        *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation; either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* FXMOVIEMANAGER is distributed in the hope that it will be useful,             *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with this program; if not, write to the Free Software                   *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA     * 
********************************************************************************/


            /****************************************************
            * Acknowledgements:                                 *
            * the code written in this file is based on ffplay: * 
            * Simple Media Player based on the FFmpeg libraries *
            * Copyright (c) 2003 Fabrice Bellard,               *
            *                                                   *
            * I found also useful a tutorial by Stephen Dranger *
            * who updated the original tutorial by Martin Bohme *
            ****************************************************/


#include "config.h"

#ifdef HAVE_INTTYPES_H
 #include <inttypes.h>
#endif

#include <math.h>
#include <limits.h>

/* utils */
#include <libavutil/avutil.h>
#include <libavutil/mathematics.h>
#include <libavutil/pixdesc.h>
#include <libavutil/imgutils.h>
#include <libavutil/dict.h>
#include <libavutil/samplefmt.h>
#include <libavutil/audioconvert.h>
#include <libavcodec/avcodec.h>
#include <libavcodec/avfft.h>

/* begin core libraries */
#include <libavformat/avformat.h>
#include <libswscale/swscale.h>
#include <libavutil/opt.h> /* AVOptions is requiried to manage bad src image pointers */
#include <libswresample/swresample.h>
/* end core libraries */

#include <SDL/SDL.h>
#include <SDL/SDL_thread.h>

#include <unistd.h>
#include <assert.h>
#include <string.h>

#ifdef __MINGW32__
 #undef main /* Prevents SDL from overriding main() */
#endif


/* SDL audio buffer size, in samples. Should be small to have precise
   Audio/Video sync as SDL does not have hardware buffer fullness info. */
#define SDL_AUDIO_BUFFER_SIZE 1024

#define MAX_TOTAL_QUEUE_SIZE (15 * 1024 * 1024)
#define MAX_AUDIO_QUEUE_SIZE (20 * 16 * 1024)
#define MAX_VIDEO_QUEUE_FRAMES 5

/* no AV sync correction is done if below the AV sync threshold */
#define AV_SYNC_THRESHOLD 0.01
/* no AV correction is done if too big error */
#define AV_NOSYNC_THRESHOLD 10.0

/* maximum audio speed change to get correct sync */
#define SAMPLE_CORRECTION_PERCENT_MAX 10
/* we use about AUDIO_DIFF_AVG_NB A-V differences to make the average */
#define AUDIO_DIFF_AVG_NB 20

/* enum events type and bind to SDL UserEvents*/
#define FF_ALLOC_EVENT   (SDL_USEREVENT)
#define FF_REFRESH_EVENT (SDL_USEREVENT + 1)
#define FF_QUIT_EVENT    (SDL_USEREVENT + 2)

#define VIDEO_PICTURE_QUEUE_SIZE 2

/* to be changed if your path is different*/
#ifdef INSTALLATIONPATH
 #define  ICON_LG  INSTALLATIONPATH "/share/fxmm/cmd_logo.bmp"
 #define  ICON_B   INSTALLATIONPATH "/share/fxmm/cmd_back.bmp"
 #define  ICON_BB  INSTALLATIONPATH "/share/fxmm/cmd_bback.bmp"
 #define  ICON_F   INSTALLATIONPATH "/share/fxmm/cmd_for.bmp"
 #define  ICON_FF  INSTALLATIONPATH "/share/fxmm/cmd_ffor.bmp"
 #define  ICON_SG  INSTALLATIONPATH "/share/fxmm/cmd_size_g.bmp"
 #define  ICON_SO  INSTALLATIONPATH "/share/fxmm/cmd_size_o.bmp"
 #define  ICON_SS  INSTALLATIONPATH "/share/fxmm/cmd_size_s.bmp"
#else
 #define  ICON_LG  "/usr/local/share/fxmm/cmd_logo.bmp"
 #define  ICON_B   "/usr/local/share/fxmm/cmd_back.bmp"
 #define  ICON_BB  "/usr/local/share/fxmm/cmd_bback.bmp"
 #define  ICON_F   "/usr/local/share/fxmm/cmd_for.bmp"
 #define  ICON_FF  "/usr/local/share/fxmm/cmd_ffor.bmp"
 #define  ICON_SG  "/usr/local/share/fxmm/cmd_size_g.bmp"
 #define  ICON_SO  "/usr/local/share/fxmm/cmd_size_o.bmp"
 #define  ICON_SS  "/usr/local/share/fxmm/cmd_size_s.bmp"
#endif

static int sws_flags = SWS_BICUBIC;

typedef struct PacketQueue {
  AVPacketList *first_pkt, *last_pkt;
  int quit;
  int nb_packets;
  int size;
  SDL_mutex *mutex;
  SDL_cond *cond;
} PacketQueue;

typedef struct VideoPicture {
  double pts;                      ///< presentation time stamp for this picture
  double duration;                 ///< expected duration of the frame
  int64_t pos;     
  SDL_Overlay *bmp;
  int width, height;               ///< source height & width
  int allocated;
  int skip;
  enum PixelFormat pix_fmt;
} VideoPicture;

typedef struct VideoState {

  AVFormatContext *avf_ctx;

  int             video_stream; 
  int             audio_stream;

  int             av_sync_type;
  double          external_clock;   ///< external clock base
  int64_t         external_clock_time;

  int             seek_req;
  int64_t         seek_pos;
  int64_t         seek_rel;

  double          audio_clock;
  AVStream        *audio_st;
  PacketQueue     audioq;
  DECLARE_ALIGNED(16,uint8_t,audio_buf2)[AVCODEC_MAX_AUDIO_FRAME_SIZE * 4];

  uint8_t         silence_buf[SDL_AUDIO_BUFFER_SIZE];
  uint8_t         *audio_buf;
  uint8_t         *audio_buf1;

  unsigned int    audio_buf_size;   ///< in bytes
  int             audio_buf_index;  ///< in bytes

  int             audio_hw_buf_size; 
  double          audio_diff_cum;   ///< used for AV difference average computation
  double          audio_diff_avg_coef;
  double          audio_diff_threshold;
  int             audio_diff_avg_count;

  AVPacket        audio_pkt;
  AVPacket        audio_pkt_tmp;

  enum AVSampleFormat audio_src_fmt;
  enum AVSampleFormat audio_tgt_fmt;
  int             audio_src_channels;
  int             audio_tgt_channels;
  int64_t         audio_src_channel_layout;
  int64_t         audio_tgt_channel_layout;
  int             audio_src_freq;
  int             audio_tgt_freq;

  double          audio_current_pts;
  double          audio_current_pts_time;
  AVFrame         *frame;

  int             frame_drops_early;
  int             frame_drops_late;

  double          frame_timer;
  double          frame_last_pts;
  double          frame_last_duration;
  double          frame_last_dropped_pts;
  double          frame_last_returned_time;
  double          frame_last_filter_delay;
  int64_t         frame_last_dropped_pos;
  double          video_clock;             ///< pts of last decoded frame / predicted pts of next decoded frame
  double          video_current_pts;       ///< current displayed pts (different from video_clock if frame fifos are used)
  double          video_current_pts_time;  ///< time (av_gettime) at which we updated video_current_pts - used to have running video pts
  int64_t         video_current_pos;   

  AVStream        *video_st;
  PacketQueue     videoq;
  VideoPicture    pictq[VIDEO_PICTURE_QUEUE_SIZE];
  int             pictq_size, pictq_rindex, pictq_windex;

  struct SwrContext *swr_ctx;
  struct SwsContext *img_convert_ctx;
  SwsFilter       *img_filter;

  SDL_mutex       *pictq_mutex;
  SDL_cond        *pictq_cond;
  SDL_Thread      *decode_tid;
  SDL_Thread      *video_tid;
  SDL_Thread      *refresh_tid;

  char            filename[1024];
  int             width;
  int             height;
  int             quit;
  int             pause_curr;
  int             pause_last;
  int             pause_back;
  int             step;
  int             refresh;

  float           luma_GBlur;
  float           chro_GBlur;
  float           luma_Sharpen;
  float           chro_Sharpen;
  float           chro_HShift;
  float           chro_VShift;

} VideoState;

enum {
  AV_SYNC_AUDIO_MASTER,
  AV_SYNC_VIDEO_MASTER,
  AV_SYNC_EXTERNAL_MASTER,
};

enum {
  IF_LUMA_GB,
  IF_CHRO_GB,
  IF_LUMA_S,
  IF_CHRO_S,
  IF_CHRO_HS,
  IF_CHRO_VS,
  IF_NOPTS,
};

static int av_sync_type = AV_SYNC_AUDIO_MASTER;
static int cycle_audio_ch = 0;

static char *imgcommand = NULL;
static int time_command = 0;
static int show_command = 0;

// options passed by the caller
static intmax_t time_s = 0;  ///< start time
static intmax_t time_d = 0;  ///< duration time
static float time_a = 0;     ///< audio delay time
static int screen_w = 0;
static int screen_h = 0;
static int full = 0;
static int loop = 0;
static int drop = 0;
static int disable_video = 0;
static int disable_audio = 0;
static int fs_screen_w = 0;
static int fs_screen_h = 0;
static int enlarge_h = 0;
static int enlarge_v = 0;

static float lumaGBlur   = 0.0;
static float chroGBlur   = 0.0;
static float lumaSharpen = 0.0;
static float chroSharpen = 0.0;
static float chroHShift  = 0.0;
static float chroVShift  = 0.0;

// the showing surface
static SDL_Surface *screen = NULL;
// since we have only one decoding thread, we can use a global
// variable instead of a thread local variable
static VideoState *global_video_state;

static AVPacket flush_pkt;
// packet queue duplicate packet for mpeg1-video header missing 
static int packet_queue_dup(AVPacket *pkt) {
  if(pkt != &flush_pkt && av_dup_packet(pkt) < 0) {
    return -1;
  }
  return 0;
}

// packet queue store packet
static int packet_queue_put(PacketQueue *q, AVPacket *pkt) {

  AVPacketList *pkt_lst;

  packet_queue_dup(pkt);
  // allocate memory
  pkt_lst = av_malloc(sizeof(AVPacketList));
  if (!pkt_lst)
    return -1;

  // populate
  pkt_lst->pkt = *pkt;
  pkt_lst->next = NULL;

  // lock
  SDL_LockMutex(q->mutex);

  if (!q->last_pkt)
    q->first_pkt = pkt_lst;
  else
    q->last_pkt->next = pkt_lst;

  // increment packet queue size
  q->last_pkt = pkt_lst;
  q->nb_packets++;
  q->size += pkt_lst->pkt.size + sizeof(*pkt_lst);

  SDL_CondSignal(q->cond);
  // unlock
  SDL_UnlockMutex(q->mutex);

  return 0;
}

// packet queue initialize
static void packet_queue_init(PacketQueue *q) {
  memset(q, 0, sizeof(PacketQueue));
  q->mutex = SDL_CreateMutex();
  q->cond = SDL_CreateCond();
  packet_queue_put(q, &flush_pkt);
}

// packet queue flush
static void packet_queue_flush(PacketQueue *q) {

  AVPacketList *pkt_lst, *pkt_lst1;
  // lock
  SDL_LockMutex(q->mutex);

  for(pkt_lst = q->first_pkt; pkt_lst != NULL; pkt_lst = pkt_lst1) {
    pkt_lst1 = pkt_lst->next;
    av_free_packet(&pkt_lst->pkt);
    av_freep(&pkt_lst);
  }
  q->last_pkt = NULL;
  q->first_pkt = NULL;
  q->nb_packets = 0;
  q->size = 0;
  // unlock
  SDL_UnlockMutex(q->mutex);
}

// packet queue destroy
static void packet_queue_destroy(PacketQueue *q) {
    packet_queue_flush(q);
    SDL_DestroyMutex(q->mutex);
    SDL_DestroyCond(q->cond);
}

// packet queue lock
static void packet_queue_quit(PacketQueue *q) {
    SDL_LockMutex(q->mutex);
    q->quit = 1;
    SDL_CondSignal(q->cond);
    SDL_UnlockMutex(q->mutex);
}

// return -1 if aborted, 0 if no packet and 1 if packet.
static int packet_queue_get(PacketQueue *q, AVPacket *pkt, int block) {

  AVPacketList *pkt_lst;
  int ret;

  SDL_LockMutex(q->mutex);
  
  for(;;) {
    if(q->quit) {
      ret = -1;
      break;
    }

    pkt_lst = q->first_pkt;
    if (pkt_lst) {
      q->first_pkt = pkt_lst->next;
      if (!q->first_pkt)
        q->last_pkt = NULL;
      q->nb_packets--;
      q->size -= pkt_lst->pkt.size + sizeof(*pkt_lst);
      *pkt = pkt_lst->pkt;
      av_free(pkt_lst);
      ret = 1;
      break; 
    } else if (!block) {
      ret = 0;
      break;
    } else {
      SDL_CondWait(q->cond, q->mutex);
    }
  }
  SDL_UnlockMutex(q->mutex);

  return ret;
}


// get the current audio clock value
static double get_audio_clock(VideoState *vs) {
    double val;
    int64_t time_av;

    if (vs->pause_curr) {
      return vs->audio_current_pts;
    } else {
      time_av = av_gettime();
      val = (double)vs->audio_current_pts_time + (double)(time_av * 1e-6);
      return val;
    }
}
// get the current video clock value
static double get_video_clock(VideoState *vs) {
    double val;
    int64_t time_av;

    if (vs->pause_curr) {
        return vs->video_current_pts;
    } else {
       time_av = av_gettime();
       val = (double)vs->video_current_pts_time + (double)(time_av * 1e-6);
       return val;
    }
}
// get the current external clock value
static double get_external_clock(VideoState *vs) {
    double val;
    int64_t time_av, time_ex ;

    time_av = av_gettime();
    time_ex = vs->external_clock_time;
    val = vs->external_clock + (double)((time_av - time_ex) * 1e-6);
    return val;
}
// get the current clock value
static double get_master_clock(VideoState *vs) {
    double val;

    if (vs->av_sync_type == AV_SYNC_VIDEO_MASTER) {
        if (vs->video_st)
            val = get_video_clock(vs);
        else if (vs->audio_st)
            val = get_audio_clock(vs);
    } else if (vs->av_sync_type == AV_SYNC_AUDIO_MASTER) {
        if (vs->audio_st)
            val = get_audio_clock(vs);
        else if (vs->video_st)
            val = get_video_clock(vs);
    } else {
        val = get_external_clock(vs);
    }
    return val;
}

static inline void fill_rectangle(SDL_Surface *screen,
                                  int x, int y, int w, int h, int color) {
    SDL_Rect rect;
    rect.x = x;
    rect.y = y;
    rect.w = w;
    rect.h = h;
    SDL_FillRect(screen, &rect, color);
}

static void close_video_picture(VideoState *vs) {

    VideoPicture *vp;
    int i;
    vs->quit = 1;

    //Wait for the threads to finish
    SDL_WaitThread(vs->decode_tid, NULL);
    SDL_WaitThread(vs->refresh_tid, NULL);

    for(i=0;i<VIDEO_PICTURE_QUEUE_SIZE; i++) {
      vp = &vs->pictq[i];

      if (vp->bmp) {
        SDL_FreeYUVOverlay(vp->bmp);
        vp->bmp = NULL;
      }
    }

    SDL_DestroyMutex(vs->pictq_mutex);
    SDL_DestroyCond(vs->pictq_cond);

    // free the sws image filter
    if(vs->img_filter)
      sws_freeFilter(vs->img_filter);

    // Frees the sws scaler context.
    if (vs->img_convert_ctx)
      sws_freeContext(vs->img_convert_ctx);

    av_free(vs);
}

static void do_exit(VideoState *vs) {
    if (vs) {
        close_video_picture(vs);
        vs = NULL;
    }
    av_lockmgr_register(NULL);
    // Close the network protocol
    avformat_network_deinit();

    SDL_Quit();
    av_log(NULL, 
           AV_LOG_QUIET, 
           "%s\n", 
           "Thank you for using fxmoviemanager");
    return;
}

// synchronize the audio stream:
// return a new audio buffer size (if found video or external master)
// deleting or adding samples to better sync with master clock
// and usefull also for passing an audiodelay value to the player
static int synchronize_audio(VideoState *vs, int nb_samples) {

  int delay_nb_samples; 
  int wanted_nb_samples = nb_samples;

  // if AUDIO is not the master clock, then we try to remove or add samples from AUDIO
  // in order to correct the clock according to the VIDEO master
  if (((vs->av_sync_type == AV_SYNC_VIDEO_MASTER && vs->video_st) ||
        vs->av_sync_type == AV_SYNC_EXTERNAL_MASTER)) {
    double diff_clock, avg_diff;
    int min_nb_samples, max_nb_samples;
    // cumuate till we have a Master > Audio (every time base)
    diff_clock =  get_audio_clock(vs) - get_master_clock(vs);
	// if the threshold is not reached 
    if(diff_clock < AV_NOSYNC_THRESHOLD) {
      // cumulate the diffs
      vs->audio_diff_cum = diff_clock + vs->audio_diff_avg_coef * vs->audio_diff_cum;
      if(vs->audio_diff_avg_count < AUDIO_DIFF_AVG_NB) {
        // not enough measures to have a correct estimate
	    vs->audio_diff_avg_count++;
      } else {
        // estimate the A-V difference
        avg_diff = vs->audio_diff_cum * (1.0 - vs->audio_diff_avg_coef);

	    if(fabs(avg_diff) >= vs->audio_diff_threshold) {
	      wanted_nb_samples = nb_samples + (int)(diff_clock * vs->audio_src_freq);
          // in case the calculation fails and audio adjustment is needed, 
          // we manually increase or decrease the wanted size 
          // so that the real sample size will be appended or cutted
          if(time_a !=0) {
            delay_nb_samples = (int)(time_a * vs->audio_src_freq);
            wanted_nb_samples = nb_samples + delay_nb_samples;
          }
          // CALCULATION: estimate the A-V difference
          min_nb_samples = ((nb_samples * (100 - SAMPLE_CORRECTION_PERCENT_MAX) / 100));
          max_nb_samples = ((nb_samples * (100 + SAMPLE_CORRECTION_PERCENT_MAX) / 100));
          wanted_nb_samples = FFMIN(FFMAX(wanted_nb_samples, min_nb_samples), max_nb_samples);
	    }  
      }
    } else {
      // the difference is too big; RESET to zero
      vs->audio_diff_avg_count = 0;
      vs->audio_diff_cum = 0;
    }
  }
  return wanted_nb_samples;
}

// decode one audio frame and returns its uncompressed size
static int audio_decode_frame(VideoState *vs, double *pts_ptr) {

  AVPacket *pkt = &vs->audio_pkt;
  AVPacket *pkt_tmp = &vs->audio_pkt_tmp;
  AVCodecContext *dec = vs->audio_st->codec;
  int64_t dec_channel_layout;
  int got_frame, or_data_size, cv_data_size, len_a, len_b, wanted_nb_samples;
  double pts;
  int new_pkt = 0;
  int flush_done = 0;

  for(;;) {
    // an audio packet can contain several frames
    while(pkt_tmp->size > 0 || (!pkt_tmp->data && new_pkt)) {
      if (!vs->frame) {
        if (!(vs->frame = avcodec_alloc_frame()))
          return AVERROR(ENOMEM);
      } else
        avcodec_get_frame_defaults(vs->frame);

      if(flush_done)
        break;
      new_pkt = 0;
      len_a = avcodec_decode_audio4(dec, vs->frame, &got_frame, pkt_tmp); 
      // if error, skip the frame
      if(len_a < 0) {
	    pkt_tmp->size = 0;
	    break;
      }

      pkt_tmp->data += len_a;
      pkt_tmp->size -= len_a;

      if(!got_frame) {
        // stop sending empty packets if the decoding is finished
        if (!pkt_tmp->data && dec->codec->capabilities & CODEC_CAP_DELAY) {
          flush_done = 1;
        }
	    // No data yet, get next frame
	    continue;
      }
      // original audio data size
      or_data_size = av_samples_get_buffer_size(NULL, 
                                                dec->channels, 
                                                vs->frame->nb_samples, 
                                                dec->sample_fmt, 
                                                1);
      cv_data_size = or_data_size;

      dec_channel_layout = (dec->channel_layout && dec->channels == av_get_channel_layout_nb_channels(dec->channel_layout)) ? 
                            dec->channel_layout : 
                            av_get_default_channel_layout(dec->channels);

      // audio synchronization
      wanted_nb_samples = synchronize_audio(vs, vs->frame->nb_samples);

      // if conversion has not be done yet try to convert
      if (dec_channel_layout != vs->audio_src_channel_layout || 
          dec->sample_fmt != vs->audio_src_fmt || 
          dec->sample_rate != vs->audio_src_freq ||
          (wanted_nb_samples != vs->frame->nb_samples && !vs->swr_ctx)){
        // free the converter buffer
        if (vs->swr_ctx)
          swr_free(&vs->swr_ctx);
        // try and use the audio converter
        vs->swr_ctx = swr_alloc_set_opts(NULL,
                                         vs->audio_tgt_channel_layout, 
                                         vs->audio_tgt_fmt, 
                                         vs->audio_tgt_freq,
                                         dec_channel_layout,
                                         dec->sample_fmt,   
                                         dec->sample_rate,
                                         0, 
                                         NULL);

        // if unable to convert audio skip setting converted variables and print to stdout
        if (!vs->swr_ctx || swr_init(vs->swr_ctx) < 0) {
          fprintf(stderr, "Unable to convert FROM sample rate %d Hz %s format %d channels TO sample rate %d Hz %s format %d channels!\n",
          dec->sample_rate,
          av_get_sample_fmt_name(dec->sample_fmt),
          dec->channels,
          vs->audio_tgt_freq,
          av_get_sample_fmt_name(vs->audio_tgt_fmt),
          vs->audio_tgt_channels);
          break; // exit the loop
        }
        // set converted variables
        vs->audio_src_channel_layout = dec_channel_layout;
        vs->audio_src_fmt = dec->sample_fmt;
        vs->audio_src_freq = dec->sample_rate;
        vs->audio_src_channels = dec->channels;
      }

      // if conversion was done
      if (vs->swr_ctx) {
        uint8_t *out[] = { vs->audio_buf2 };
        const uint8_t *in[] = { vs->frame->data[0] };

        if (wanted_nb_samples != vs->frame->nb_samples) {
          if (swr_set_compensation(vs->swr_ctx, (wanted_nb_samples - vs->frame->nb_samples) * vs->audio_tgt_freq / dec->sample_rate,
                                   wanted_nb_samples * vs->audio_tgt_freq / dec->sample_rate) < 0) {
               fprintf(stderr, "Compensation failed\n");
               break;
          }
        }

        // calculate converted audio buffer
        len_b = swr_convert(vs->swr_ctx, 
                            out, 
                            sizeof(vs->audio_buf2) / vs->audio_tgt_channels / av_get_bytes_per_sample(vs->audio_tgt_fmt),
                            in, 
                            vs->frame->nb_samples);

        if (len_b < 0) {
          fprintf(stderr, "Audio convesion failed\n");
          break; // exit the loop
        }
        if (len_b == sizeof(vs->audio_buf2) / vs->audio_tgt_channels / av_get_bytes_per_sample(vs->audio_tgt_fmt)) {
          swr_init(vs->swr_ctx);
          fprintf(stderr, "Not enough audio buffer\n");
        }
        // converted audio data size
        cv_data_size = len_b * vs->audio_tgt_channels * av_get_bytes_per_sample(vs->audio_tgt_fmt);
        vs->audio_buf = vs->audio_buf2;
      } else {
        vs->audio_buf = vs->frame->data[0];
      }

      // return a pointer to the current audio clock time
      pts = vs->audio_clock;
      *pts_ptr = pts;
      // increment audio clock time with a sample format step 
      vs->audio_clock += (double)or_data_size / (dec->channels * dec->sample_rate * av_get_bytes_per_sample(dec->sample_fmt));

      // return converted audio sample size and continue with the next frame of the audio packet
      return cv_data_size;
    }

    // free the current packet
    if(pkt->data)
      av_free_packet(pkt);

    if(vs->pause_curr || vs->audioq.quit) {
      return -1;
    }

    // get next packet on quewe
    if ((new_pkt = packet_queue_get(&vs->audioq, pkt, 1)) < 0)
      return -1;

    if(pkt->data == flush_pkt.data)
      avcodec_flush_buffers(dec);

    *pkt_tmp = *pkt;  

    // if got new packet update the audio clock 
    if(pkt->pts != AV_NOPTS_VALUE) {
      vs->audio_clock = av_q2d(vs->audio_st->time_base)*pkt->pts;
    }
  }
}

// prepare a new SDL audio buffer  (substitute  current_audio_buf)
static void sdl_audio_callback(void *userdata, Uint8 *stream, int len) {
  VideoState *vs = userdata;
  int audio_size, len_a, buffer, bytes_per_sec=0;
  double pts; 
  int frame_size = av_samples_get_buffer_size(NULL, 
                                              vs->audio_tgt_channels, 
                                              1, 
                                              vs->audio_tgt_fmt, 
                                              1);
  int64_t time_curr = av_gettime();

  while(len > 0) {
    if(vs->audio_buf_index >= vs->audio_buf_size) {

       audio_size = audio_decode_frame(vs, &pts);
       if(audio_size < 0) {
	      // if error the output silence
	      vs->audio_buf = vs->silence_buf;
	      vs->audio_buf_size = sizeof(vs->silence_buf) / frame_size * frame_size;
       } else {
	      vs->audio_buf_size = audio_size;
       }

      vs->audio_buf_index = 0;
    }

    len_a = vs->audio_buf_size - vs->audio_buf_index;

    if(len_a > len)
       len_a = len;

    memcpy(stream, (uint8_t *)vs->audio_buf + vs->audio_buf_index, len_a);

    len -= len_a;
    stream += len_a;
    vs->audio_buf_index += len_a;
  }
  // calculate the audio stream bytes per second
  bytes_per_sec = vs->audio_tgt_freq * vs->audio_tgt_channels * av_get_bytes_per_sample(vs->audio_tgt_fmt);

  buffer = vs->audio_buf_size - vs->audio_buf_index;
  // Assume the audio driver that is used by SDL has two periods
  vs->audio_current_pts = vs->audio_clock - (double)(2 * vs->audio_hw_buf_size + buffer) / bytes_per_sec;

  vs->audio_current_pts_time = vs->audio_current_pts - (double)(time_curr * 1e-6);
}

// set the SDL screen
static int video_screen_open(VideoState *vs){
    int sdl_flags = SDL_HWSURFACE|SDL_ASYNCBLIT|SDL_HWACCEL;
    int w, h, sdl_bpp;
    int wc, hc;

    if(full==1) sdl_flags |= SDL_FULLSCREEN;
    else        sdl_flags |= SDL_RESIZABLE;

    if (full==1 && fs_screen_w) {
        w = fs_screen_w;
        h = fs_screen_h;
    } else if(full==0 && screen_w){
        w = screen_w;
        h = screen_h;
    }else if (vs->video_st && vs->video_st->codec->width){
        w = vs->video_st->codec->width;
        h = vs->video_st->codec->height;
    } else {
        w = 720;
        h = 400;
    }

    if(screen &&
       vs->width == screen->w &&
       screen->w == wc &&
       vs->height == screen->h &&
       screen->h == hc)
          return 0;

#if HAVE_STRUCT_SDL_VIDEOINFOP
      const SDL_VideoInfo *sdl_vi = SDL_GetVideoInfo();
      switch (sdl_vi->vfmt->BitsPerPixel) {
          case 16:
          case 24:
          case 32:
              sdl_bpp = sdl_vi->vfmt->BitsPerPixel;
              break;
          default:
              sdl_bpp = 16;
              break;
      }
#else
      sdl_bpp = 0;
#endif

    // enlarge the horizontal size of the picture if requested
    switch (enlarge_h) { 
      case 0: // original size
        wc = w;
        break;
      case 1:
        wc = w + w*0.1;
        break;
      case 2:
        wc = w + w*0.2;
        break;
      case 3:
        wc = w + w*0.3;
        break;
      case 4:
        wc = w + w*0.4;
        break;
      case 5:
        wc = w + w*0.5;
        break;
      case 6:
        wc = w + w*0.6; // max size
        break;
    }
    // enlarge the vertical size of the picture if requested
    switch (enlarge_v) { 
      case 0: // original size
        hc = h;
        break;
      case 1:
        hc = h + h*0.1;
        break;
      case 2:
        hc = h + h*0.2;
        break;
      case 3:
        hc = h + h*0.3;
        break;
      case 4:
        hc = h + h*0.4;
        break;
      case 5:
        hc = h + h*0.5;
        break;
      case 6:
        hc = h + h*0.6; // max size
        break;
    }
  // set the window icon
  SDL_Surface *bitmap;
  bitmap = SDL_LoadBMP(ICON_LG);
  if (bitmap == NULL) {
    fprintf(stderr, "Unable to load %s: %s\n", ICON_LG, SDL_GetError());
    return;
  }
  SDL_WM_SetIcon(bitmap,NULL);

#ifndef __APPLE__
    screen = SDL_SetVideoMode(wc, hc, sdl_bpp, sdl_flags);
#else
    // setting bits_per_pixel = 0 or 32 causes blank video on OS X 
    screen = SDL_SetVideoMode(wc, hc, 24, sdl_flags);
#endif
    if (!screen) {
        fprintf(stderr, "SDL: Unable to set video mode -> quitting\n");
        do_exit(vs);
    }
    //set the window caption
    char cap_title[1024]="FXMM: ";
    strncat(cap_title, vs->filename, sizeof(cap_title));
    SDL_WM_SetCaption( cap_title, NULL );

    // store the setting
    vs->width = screen->w;
    vs->height = screen->h;

    return 0;
}

// display the command given
static void video_command_display() {
  SDL_Surface *bitmap;
  SDL_Rect rect;

  bitmap = SDL_LoadBMP(imgcommand);
  if (bitmap == NULL) {
    fprintf(stderr, "Unable to load %s: %s\n", imgcommand, SDL_GetError());
    return;
  }

  rect.x = 0;
  rect.y = 0;
  rect.w = bitmap->w;
  rect.h = bitmap->h;

  //SDL_LockSurface(screen);
  SDL_BlitSurface(bitmap, NULL, screen, &rect);
  SDL_UpdateRects(screen, 1, &rect);
  //SDL_UnlockSurface(screen);
}

// display the current picture, if any
static void video_image_display(VideoState *vs) {
  VideoPicture *vp;
  AVPicture pict;
  SDL_Rect rect;
  float aspect_ratio;
  int w, h, x, y;
  int i;

  vp = &vs->pictq[vs->pictq_rindex];
  if(vp->bmp) {
    // check if on AVStream is present the picture aspect ratio
    if (vs->video_st->sample_aspect_ratio.num > 0) {
      aspect_ratio = av_q2d(vs->video_st->sample_aspect_ratio); 
    } else if (vs->video_st->codec->sample_aspect_ratio.num > 0) {
      aspect_ratio = av_q2d(vs->video_st->codec->sample_aspect_ratio); 
    } else {
      aspect_ratio = 1;
    }

    // if not present we get it from VideoPicture
    aspect_ratio *= (float)vp->width / (float)vp->height;
    //fprintf(stderr, "aspect ratio:%f \n", aspect_ratio);

    // try and display a picture size keeping a good aspect ratio
    // basing on the assumption that height is the best measure
    h = vs->height;
    w = ((int)rint(h * aspect_ratio)) & ~1;
    // if it fails basing on the assumption that width is the best measure
    if(w > vs->width) {
      w = vs->width;
      h = ((int)rint(w / aspect_ratio)) & ~1;
    }

    rect.w = FFMAX(w, 1);
    rect.h = FFMAX(h, 1);

    // calculate the start point coordinates for the picture
    x = (vs->width - w) / 2;
    y = (vs->height - h) / 2;

    rect.x = x;
    rect.y = y;


    // display the window
    SDL_LockYUVOverlay(vp->bmp);
    SDL_GetClipRect(screen, &rect);
    SDL_DisplayYUVOverlay(vp->bmp, &rect);

    if (show_command == 1 && time_command <= 30) {
      time_command += 1;
      //fprintf(stderr, "time:%i \n", time_command);
      video_command_display();
    } else {
      time_command = 0; 
      show_command = 0;
    }

    SDL_UnlockYUVOverlay(vp->bmp);
  }

}

// call subroutines for video display
static void video_display(VideoState *vs) {

    if(!screen)
       video_screen_open(vs);

    if(vs->video_st) {
       video_image_display(vs);
    } else if (vs->audio_st) {
       int fgcolor = SDL_MapRGB(screen->format, 0xff, 0xff, 0xff);
       fill_rectangle(screen, 0, 0, screen->w, screen->h, fgcolor);
    }
}

// the SDL refresh clock
static int refresh_thread(void *opaque) {

    VideoState *vs = opaque;
    while(!vs->quit){
        SDL_Event event;
        event.type = FF_REFRESH_EVENT;
        event.user.data1 = opaque;
        if(!vs->refresh){
            vs->refresh = 1;
            SDL_PushEvent(&event);
        }
        usleep(vs->audio_st && vs->video_st);
     }
    return 0;
}

// synchronize the video stream
// return a new video buffer size (if found AUDIO or EXTERNAL MASTER)
// deleting or adding samples to better sync with master clock
static double synchronize_video(double delay, VideoState *vs) {

    double sync_threshold, diff_clock;

    // if VIDEO is not the master clock, then we try 
    // to remove or duplicate samples from VIDEO stream
    // in order to correct the clock according to the AUDIO master
    if (((vs->av_sync_type == AV_SYNC_AUDIO_MASTER && vs->audio_st) ||
         vs->av_sync_type == AV_SYNC_EXTERNAL_MASTER)) {
        // cumuate
        diff_clock = get_video_clock(vs) - get_master_clock(vs);

        sync_threshold = FFMAX(AV_SYNC_THRESHOLD, delay);

        // SKIP or REPEAT video frame if in the used range
        if (fabs(diff_clock) < AV_NOSYNC_THRESHOLD) {
            if (diff_clock <= -sync_threshold) {
                delay = 0;
            }
            else if (diff_clock >= sync_threshold) {
                delay = 2 * delay;
            }
        }

        // in case the difference is too big for automatic video synchro, 
        // we can use the AUDIO_DELAY option to correct this difference.
        if(time_a != 0) {
          delay -= (double)(time_a);
        }
    }

    av_dlog(NULL,
            "video: delay=%0.3f A-V=%f\n",
            delay, -diff_clock);

    return delay;
}

// update queue size and signal for next picture
static void pictq_next_picture(VideoState *vs) {

    if (++vs->pictq_rindex == VIDEO_PICTURE_QUEUE_SIZE)
       vs->pictq_rindex = 0;

    SDL_LockMutex(vs->pictq_mutex);
    vs->pictq_size--;
    SDL_CondSignal(vs->pictq_cond);
    SDL_UnlockMutex(vs->pictq_mutex);
}

// update current video pts
static void update_video_pts(VideoState *vs, double vppts, int64_t vppos) {

    double time_curr = av_gettime() * 1e-6;
    vs->video_current_pts = vppts;
    vs->video_current_pts_time = vs->video_current_pts - time_curr;
    vs->video_current_pos = vppos;
    vs->frame_last_pts = vppts;
}

static void video_refresh_timer(void *opaque) {

  VideoState *vs = opaque;
  VideoPicture *vp;

  // VIDEO STREAM 
  if(vs->video_st) {
tryagain:
     if(vs->pictq_size == 0) {
        SDL_LockMutex(vs->pictq_mutex);
        if(vs->frame_last_dropped_pts != AV_NOPTS_VALUE && vs->frame_last_dropped_pts > vs->frame_last_pts) {
           update_video_pts(vs, vs->frame_last_dropped_pts, vs->frame_last_dropped_pos);
           vs->frame_last_dropped_pts = AV_NOPTS_VALUE;
        }
        SDL_UnlockMutex(vs->pictq_mutex);
     } else {
        double last_duration, duration, time_curr, delay;
        // dequeue the picture
        vp = &vs->pictq[vs->pictq_rindex];

        if (vp->skip) {
           pictq_next_picture(vs);
           goto tryagain;
        }

      // compute nominal last_duration
      last_duration = vp->pts - vs->frame_last_pts;
      if (last_duration > 0 && last_duration < 10.0) {
         // if found a suitable duration of the last frame, 
         // update last_duration in video state
         vs->frame_last_duration = last_duration;
      }
      // call to sync the video stream with the master one
      delay = synchronize_video(vs->frame_last_duration, vs);

      time_curr = av_gettime() * 1e-6;
      if(time_curr < vs->frame_timer + delay)
         return;

      if (delay > 0)
         vs->frame_timer += delay * FFMAX(1, floor((time_curr - vs->frame_timer) / delay));

      SDL_LockMutex(vs->pictq_mutex);
      update_video_pts(vs, vp->pts, vp->pos);
      SDL_UnlockMutex(vs->pictq_mutex);

      if(vs->pictq_size > 1) {
         VideoPicture *nextvp = &vs->pictq[(vs->pictq_rindex+1)%VIDEO_PICTURE_QUEUE_SIZE];
         duration = nextvp->pts - vp->pts; // More accurate this way, 1/time_base is often not reflecting FPS
      } else {
         duration = vp->duration;
      }
      // if video frame drop
      if((drop==1 || (drop==0 && vs->audio_st)) && time_curr > vs->frame_timer + duration){
         if(vs->pictq_size > 1){
            vs->frame_drops_late++;
            pictq_next_picture(vs);
            goto tryagain;
         }
      }

      // show the picture
      if(disable_video == 0)
         video_display(vs);
      
      pictq_next_picture(vs);
    }
  // AUDIO STREAM
  } else if (vs->audio_st) {
     // display a rectangle
     if (disable_video == 0)
        video_display(vs);
  }
}

// allocate a picture 
// it needs to be done in main thread to avoid potential locking problems
static void alloc_picture(void *opaque) {

  VideoState *vs = opaque;
  VideoPicture *vp;

  vp = &vs->pictq[vs->pictq_windex];

  if(vp->bmp) {
    SDL_FreeYUVOverlay(vp->bmp);
  }

  vp->width = vs->video_st->codec->width;
  vp->height = vs->video_st->codec->height;
  vp->pix_fmt = vs->video_st->codec->pix_fmt;

  // Allocate a place to put our YUV image on that screen
  vp->bmp = SDL_CreateYUVOverlay(vp->width,
                                 vp->height,
                                 SDL_YV12_OVERLAY,
                                 screen);

  if (!vp->bmp || vp->bmp->pitches[0] < vp->width) {
    fprintf(stderr, "Error: the video system does not support an image size of %dx%d pixels\n", vp->width, vp->height );
    do_exit(vs);
  }
  
  SDL_LockMutex(vs->pictq_mutex);
  vp->allocated = 1;
  SDL_CondSignal(vs->pictq_cond);
  SDL_UnlockMutex(vs->pictq_mutex);
}

static int queue_picture(VideoState *vs, AVFrame *src_frame, double pts_o, int64_t pos) {

  VideoPicture *vp;
  double frame_delay;
  double pts = pts_o;

  // set video clock to pts
  if (pts != 0) {
    vs->video_clock = pts;
  } else {
    pts = vs->video_clock;
  }

  frame_delay = av_q2d(vs->video_st->codec->time_base);
  // for MPEG2 format, the frame can be repeated, 
  // the video clock is appended accordingly
  frame_delay += src_frame->repeat_pict * (frame_delay * 0.5);
  // update video clock for next frame
  vs->video_clock += frame_delay;

  // wait until there is space for a new picture
  SDL_LockMutex(vs->pictq_mutex);
  while(vs->pictq_size >= VIDEO_PICTURE_QUEUE_SIZE && !vs->videoq.quit) {
    SDL_CondWait(vs->pictq_cond, vs->pictq_mutex);
  }
  SDL_UnlockMutex(vs->pictq_mutex);

  if(vs->videoq.quit)
    return -1;

  // windex is set to 0 initially
  vp = &vs->pictq[vs->pictq_windex];
  vp->duration = frame_delay;

  // if NOT FRAME: allocate or resize hardware picture buffer
  if(!vp->bmp || 
     vp->width != vs->video_st->codec->width ||
     vp->height != vs->video_st->codec->height) {

    SDL_Event event;
    vp->allocated = 0;
    // the allocation must be done in the main thread to avoid locking problems
    event.type = FF_ALLOC_EVENT;
    event.user.data1 = vs;
    SDL_PushEvent(&event);

    // wait until there is a picture allocated
    SDL_LockMutex(vs->pictq_mutex);
    while(!vp->allocated && !vs->videoq.quit) {
      SDL_CondWait(vs->pictq_cond, vs->pictq_mutex);
    }
    // if the queue is aborted, we have to pop the pending ALLOC event or wait for the allocation to complete 
    if(vs->videoq.quit && SDL_PeepEvents(&event, 1, SDL_GETEVENT, SDL_EVENTMASK(FF_ALLOC_EVENT)) != 1) {
      while (!vp->allocated) {
        SDL_CondWait(vs->pictq_cond, vs->pictq_mutex);
      }
    }
    SDL_UnlockMutex(vs->pictq_mutex);

    if(vs->videoq.quit) {
      return -1;
    }
  }

  // if FRAME: if it is not skipped display it
  if(vp->bmp) {

    AVPicture pict;
    // get a pointer on the bitmap
    SDL_LockYUVOverlay(vp->bmp);

    // point pict at the queue
    memset(&pict,0,sizeof(AVPicture));

    pict.data[0] = vp->bmp->pixels[0];
    pict.data[1] = vp->bmp->pixels[2];
    pict.data[2] = vp->bmp->pixels[1];
    
    pict.linesize[0] = vp->bmp->pitches[0];
    pict.linesize[1] = vp->bmp->pitches[2];
    pict.linesize[2] = vp->bmp->pitches[1];

    // Checks if sws context can be reused, otherwise reallocates a new one.
    vs->img_convert_ctx = sws_getCachedContext(vs->img_convert_ctx,
                                               vp->width, vp->height, vp->pix_fmt,      // source format
                                               vp->width, vp->height, PIX_FMT_YUV420P,  // destination
                                               sws_flags, 
                                               vs->img_filter, vs->img_filter,          // img filters
                                               NULL);

    // If sws context is NULL
    if(vs->img_convert_ctx == NULL) {
	  fprintf(stderr,"Unable to initialize the image conversion context!\n");
	  return 1;
    }

    // Scales the image slice in src_Slice and puts the resulting scaled
    // slice in the image in dst (a slice is a sequence of consecutive rows in an image)
    sws_scale(vs->img_convert_ctx, 
              (const uint8_t * const*)src_frame->data, 
              src_frame->linesize,
	          0, 
              vp->height, 
              pict.data, 
              pict.linesize);

    // update the bitmap content 
    SDL_UnlockYUVOverlay(vp->bmp);

    vp->pts = pts;
    vp->pos = pos;
    vp->skip = 0;

    // update the picture queue index
    if(++vs->pictq_windex == VIDEO_PICTURE_QUEUE_SIZE) {
      vs->pictq_windex = 0;
    }

    SDL_LockMutex(vs->pictq_mutex);
    vs->pictq_size++;
    SDL_UnlockMutex(vs->pictq_mutex);
  }
  return 0;
}

static int get_video_frame(VideoState *vs, AVFrame *frame, int64_t *pts, AVPacket *pkt) {

    int got_picture, i;

    if (packet_queue_get(&vs->videoq, pkt, 1) < 0)
        return -1;

    if (pkt->data == flush_pkt.data) {
        avcodec_flush_buffers(vs->video_st->codec);

        SDL_LockMutex(vs->pictq_mutex);
        // Make sure there are no long delay timers 
        // ideally we should just flush the que but that's harder
        for (i=0; i<VIDEO_PICTURE_QUEUE_SIZE; i++) {
            vs->pictq[i].skip = 1;
        }
        while (vs->pictq_size && !vs->videoq.quit) {
            SDL_CondWait(vs->pictq_cond, vs->pictq_mutex);
        }
        vs->video_current_pos = -1;
        vs->frame_last_pts = AV_NOPTS_VALUE;
        vs->frame_last_duration = 0;
        vs->frame_timer = (double)av_gettime() * 1e-6;
        vs->frame_last_dropped_pts = AV_NOPTS_VALUE;
        SDL_UnlockMutex(vs->pictq_mutex);

        return 0;
    }

    avcodec_decode_video2(vs->video_st->codec, frame, &got_picture, pkt);

    if (got_picture) {
        int ret = 1;  // set error TRUE

        // use AVOptions to get a pointer (frame) to the requested field (best_e) in the class (avcodec_g).
        *pts = *(int64_t*)av_opt_ptr(avcodec_get_frame_class(), frame, "best_effort_timestamp");
        if(!*pts) 
           *pts = frame->pkt_dts;

        if(*pts == AV_NOPTS_VALUE) {
           *pts = 0;
        }

        if ( ((vs->av_sync_type == AV_SYNC_AUDIO_MASTER && vs->audio_st) || vs->av_sync_type == AV_SYNC_EXTERNAL_MASTER) &&
             (drop == 1 || (drop == 0 && vs->audio_st)) ) {
            SDL_LockMutex(vs->pictq_mutex);
            if (vs->frame_last_pts != AV_NOPTS_VALUE && *pts) {
                //fprintf(stderr, "pts:%d \n", *pts);
                //fprintf(stderr, "frame_last_pts:%f \n", vs->frame_last_pts);
                double diff_clock = get_video_clock(vs) - get_master_clock(vs);
                double dpts = av_q2d(vs->video_st->time_base) * *pts;
                double ptsdiff = dpts - vs->frame_last_pts;
                if (fabs(diff_clock) < AV_NOSYNC_THRESHOLD &&
                     ptsdiff > 0 && ptsdiff < AV_NOSYNC_THRESHOLD &&
                     diff_clock + ptsdiff - vs->frame_last_filter_delay < 0) {
                    vs->frame_last_dropped_pos = pkt->pos;
                    vs->frame_last_dropped_pts = dpts;
                    vs->frame_drops_early++;
                    ret = 0; // error FALSE
                }
            }
            SDL_UnlockMutex(vs->pictq_mutex);
        } 
        if(ret)
           vs->frame_last_returned_time = av_gettime() * 1e-6;

        return ret;
    }
    return 0;
}

static void do_pause(VideoState *vs) {

    if (vs->pause_curr) {
        double time_curr;
        time_curr = (double)(av_gettime() * 1e-6);

        vs->frame_timer += vs->video_current_pts_time + time_curr - vs->video_current_pts;

        if(vs->pause_back != AVERROR(ENOSYS)){
            vs->video_current_pts = vs->video_current_pts_time + time_curr;
        }
        vs->video_current_pts_time = vs->video_current_pts - time_curr;
    }

    vs->pause_curr = !vs->pause_curr;
    vs->step = 0;
}


// These are called whenever we allocate a frame buffer. 
// We use this to store the global_pts in a frame at the time it is allocated.
static int default_get_buffer(struct AVCodecContext *cod_ctx, AVFrame *pic) {
  uint64_t *pts = av_malloc(sizeof(uint64_t));
  *pts = AV_NOPTS_VALUE;
  pic->opaque = pts;

  int ret = avcodec_default_get_buffer(cod_ctx, pic);
  return ret;
}

static int default_reget_buffer(struct AVCodecContext *cod_ctx, AVFrame *pic) {
  avcodec_default_reget_buffer(cod_ctx, pic);
}

static void default_release_buffer(struct AVCodecContext *cod_ctx, AVFrame *pic) {
  if(pic) 
     av_freep(&pic->opaque);
  // set end of buffer to 0 for ensuring that no overreading happens for damaged mpeg streams
  memset(pic->data, 0, sizeof(pic->data));
  avcodec_default_release_buffer(cod_ctx, pic);
}

static void default_init_buffer(struct AVCodecContext *cod_ctx) {
  cod_ctx->get_buffer     = default_get_buffer;
  cod_ctx->release_buffer = default_release_buffer;
  cod_ctx->reget_buffer   = default_reget_buffer;
  cod_ctx->thread_safe_callbacks = 1;
}


static int video_thread(void *opaque) {
  VideoState *vs = opaque;
  AVFrame *p_frame = avcodec_alloc_frame();
  AVPacket pkt;
  int64_t pts_int = AV_NOPTS_VALUE;
  int64_t pos = -1;
  double pts;
  int ret;

  for(;;) {
    while (vs->pause_curr && !vs->videoq.quit)
       SDL_Delay(10);
       // get the frame of the video (ret = 0 if OK or 1 if NOT)
       ret = get_video_frame(vs, p_frame, &pts_int, &pkt);
       pos = pkt.pos;
       av_free_packet(&pkt);

       if (ret < 0) {
          av_free(p_frame);
          return 0;
       }

       vs->frame_last_filter_delay = (double)(av_gettime() * 1e-6) - vs->frame_last_returned_time;
       if (fabs(vs->frame_last_filter_delay) > AV_NOSYNC_THRESHOLD / 10.0)
          vs->frame_last_filter_delay = 0;

       pts = pts_int * av_q2d(vs->video_st->time_base);
       // put the videoframe on the queue (ret = 0 if OK or -1 if NOT)
       ret = queue_picture(vs, p_frame, pts, pos);

       if (ret < 0) {
          av_free(p_frame);
          return 0;
       }

       if (vs->step)
          do_pause(vs);
  }
  av_free(p_frame);

  return 0;
}


// open a given Audio or Video stream. Return 0 if opened without errors
static int stream_open(VideoState *vs, int stream_index) {

  AVFormatContext *avf_ctx = vs->avf_ctx;
  AVCodecContext *avc_ctx;
  AVCodec *codec;
  SDL_AudioSpec sdl_audio, sdl_spec;
  int64_t req_channel_layout = 0;
  int req_nb_channels;
  const char *sdl_env;

  if(stream_index < 0 || stream_index >= avf_ctx->nb_streams) {
     return -1;
  }
  // get a pointer to the stream codec context
  avc_ctx = avf_ctx->streams[stream_index]->codec;
  // find a decoder for the stream
  codec = avcodec_find_decoder(avc_ctx->codec_id);
  if(!codec) {
     fprintf(stderr, "Unable to find decoder!\n");
     return -1;
  }
  if (!codec || avcodec_open2(avc_ctx, codec, NULL) < 0) {
     fprintf(stderr, "Unsupported codec!\n");
     return -1;
  }

  // if DROP we deal with LOW RESOURCES values 0:1
  if(avc_ctx->codec_type == AVMEDIA_TYPE_VIDEO && drop == 1){
     avc_ctx->lowres = drop;
     if(avc_ctx->lowres) {
        avc_ctx->flags |= CODEC_FLAG_EMU_EDGE;
        fprintf(stderr, "Dealing with LOW RESOURCES!\n");
     }
     if(avc_ctx->lowres > codec->max_lowres) {
        av_log(avc_ctx, 
               AV_LOG_WARNING, 
               "The maximum LOW RESOURCES value supported by the decoder is %d\n",
                codec->max_lowres);
        avc_ctx->lowres = codec->max_lowres;
     }
  }

  if(codec->capabilities & CODEC_CAP_DR1) 
     avc_ctx->flags |= CODEC_FLAG_EMU_EDGE;

  // find the number of audio channels
  if (avc_ctx->codec_type == AVMEDIA_TYPE_AUDIO) {
     sdl_env = SDL_getenv("SDL_AUDIO_CHANNELS");
     if (sdl_env)
        req_channel_layout = av_get_default_channel_layout(SDL_atoi(sdl_env));
     if(!req_channel_layout){
        req_channel_layout = (avc_ctx->channel_layout && avc_ctx->channels == av_get_channel_layout_nb_channels(avc_ctx->channels)) ? avc_ctx->channel_layout : av_get_default_channel_layout(avc_ctx->channels);
        req_channel_layout &= ~AV_CH_LAYOUT_STEREO_DOWNMIX;
        req_nb_channels = av_get_channel_layout_nb_channels(req_channel_layout);
        // at present SDL only supports 1, 2, 4, 6 channels.
        while (req_nb_channels > 0 && (req_nb_channels == 3 || req_nb_channels == 5 || req_nb_channels > 6)) {
            req_nb_channels--;
            req_channel_layout = av_get_default_channel_layout(req_nb_channels);
        }
     }

     sdl_audio.channels = av_get_channel_layout_nb_channels(req_channel_layout);
     sdl_audio.freq = avc_ctx->sample_rate;
     // control audio sample rate
     if (sdl_audio.freq <= 0 || sdl_audio.channels <= 0) {
        fprintf(stderr, "Invalid sample rate or channel count!\n");
        return -1;
     }
  }

  // prepare SDL audio output from av codec info
  if(avc_ctx->codec_type == AVMEDIA_TYPE_AUDIO) {
     sdl_audio.format = AUDIO_S16SYS;
     sdl_audio.samples = SDL_AUDIO_BUFFER_SIZE;
     sdl_audio.callback = sdl_audio_callback;
     sdl_audio.userdata = vs;
     sdl_audio.silence = 0;
     // open SDL audio
     if(SDL_OpenAudio(&sdl_audio, &sdl_spec) < 0) {
        fprintf(stderr, "SDL_OpenAudio: %s\n", SDL_GetError());
        return -1;
     }
     vs->audio_hw_buf_size = sdl_spec.size;

     if (sdl_spec.format != AUDIO_S16SYS) {
        fprintf(stderr, "SDL advised audio format %d is not supported!\n", sdl_spec.format);
        return -1;
     }
     if (sdl_spec.channels != sdl_audio.channels) {
        req_channel_layout = av_get_default_channel_layout(sdl_spec.channels);
        if (!req_channel_layout) {
           fprintf(stderr, "SDL advised channel count %d is not supported!\n", sdl_spec.channels);
           return -1;
        }
     }
     vs->audio_src_fmt = vs->audio_tgt_fmt = AV_SAMPLE_FMT_S16;
     vs->audio_src_freq = vs->audio_tgt_freq = sdl_spec.freq;
     vs->audio_src_channel_layout = vs->audio_tgt_channel_layout = req_channel_layout;
     vs->audio_src_channels = vs->audio_tgt_channels = sdl_spec.channels;
  }

  avf_ctx->streams[stream_index]->discard = AVDISCARD_DEFAULT;

  switch(avc_ctx->codec_type) {
  case AVMEDIA_TYPE_AUDIO:
    vs->audio_stream = stream_index;
    vs->audio_st = avf_ctx->streams[stream_index];
    vs->audio_buf_size = 0;
    vs->audio_buf_index = 0;
    // averaging filter for audio syncronization
    vs->audio_diff_avg_coef = exp(log(0.01) / AUDIO_DIFF_AVG_NB);
    vs->audio_diff_avg_count = 0;
    // the audio correction is made only if larger error than this threshold
    // because there isn't enough precise audio fifo
    vs->audio_diff_threshold = 2.0 * SDL_AUDIO_BUFFER_SIZE / sdl_audio.freq;

    memset(&vs->audio_pkt, 0, sizeof(vs->audio_pkt));
    packet_queue_init(&vs->audioq);
    SDL_PauseAudio(0);
    break;
  case AVMEDIA_TYPE_VIDEO:
    vs->video_stream = stream_index;
    vs->video_st = avf_ctx->streams[stream_index];

    // initialize picture buffer
    default_init_buffer;
    // initialize picture queue
    packet_queue_init(&vs->videoq);
    // call to create the SDL video thread
    vs->video_tid = SDL_CreateThread(video_thread, vs);

    break;
  default:
    break;
  }
  return 0;
}

// safely close all streams
static void stream_close(VideoState *vs, int stream_index)
{
    AVFormatContext *avf_ctx = vs->avf_ctx;
    AVCodecContext *avctx;

    if (stream_index < 0 || stream_index >= avf_ctx->nb_streams)
        return;
    avctx = avf_ctx->streams[stream_index]->codec;

    switch(avctx->codec_type) {
    case AVMEDIA_TYPE_AUDIO:
        packet_queue_quit(&vs->audioq);
        SDL_CloseAudio();
        packet_queue_destroy(&vs->audioq);

        if (vs->swr_ctx)
            swr_free(&vs->swr_ctx);
        av_free_packet(&vs->audio_pkt);
        av_freep(&vs->audio_buf1);
        vs->audio_buf = NULL;
        av_freep(&vs->frame);

        break;
    case AVMEDIA_TYPE_VIDEO:
        packet_queue_quit(&vs->videoq);
        // NOTE: we also signal this mutex to make sure we deblock the video thread in all cases
        SDL_LockMutex(vs->pictq_mutex);
        SDL_CondSignal(vs->pictq_cond);
        SDL_UnlockMutex(vs->pictq_mutex);
        //Wait for the thread to finish
        SDL_WaitThread(vs->video_tid, NULL);
        packet_queue_destroy(&vs->videoq);
        break;
    default:
        break;
    }

    avf_ctx->streams[stream_index]->discard = AVDISCARD_ALL;
    avcodec_close(avctx);
    switch(avctx->codec_type) {
    case AVMEDIA_TYPE_AUDIO:
        vs->audio_st = NULL;
        vs->audio_stream = -1;
        break;
    case AVMEDIA_TYPE_VIDEO:
        vs->video_st = NULL;
        vs->video_stream = -1;
        break;
    default:
        break;
    }

}

void stream_seek(VideoState *vs, int64_t pos, int64_t rel) {

  if(!vs->seek_req) {
    vs->seek_pos = pos;
    vs->seek_rel = rel;
    vs->seek_req = 1;
  }
}

static int decode_interrupt_cb(void *ctx) {
  return (global_video_state && global_video_state->quit);
}

// this thread gets the stream passed by SDL
static int decode_thread(void *opaque) {

  VideoState *vs = opaque;
  AVFormatContext *avf_ctx = NULL;
  AVCodecContext *avctx;
  AVPacket pkt1, *packet = &pkt1;

  int video_index = -1;
  int audio_index = -1;
  int pkt_user_range = 0;
  int i, ret;
  int eof = 0;

  vs->video_stream=-1;
  vs->audio_stream=-1;

  global_video_state = vs;

  // This will interrupt blocking functions when we quit!
  avf_ctx = avformat_alloc_context();
  avf_ctx->interrupt_callback.callback = decode_interrupt_cb;

  // Open video file
  if(avformat_open_input(&avf_ctx, vs->filename, NULL, NULL) != 0) {
     ret = -1;
     goto fail; // Couldn't open file
  }

  vs->avf_ctx = avf_ctx;

  // Retrieve stream information
  if(avformat_find_stream_info(avf_ctx, NULL) < 0)
     goto fail; // Couldn't find stream information

  //av_dump_format(avf_ctx, 0, vs->filename, 0);
  //avf_ctx->iformat->flags & AVFMT_TS_DISCONT;

  // if start time value is given
  if (time_s != 0) {
     int64_t time_initial;
     time_initial = time_s;
     // Add the stream starttime
     if (avf_ctx->start_time != AV_NOPTS_VALUE)
        time_initial += avf_ctx->start_time;

     if (avformat_seek_file(avf_ctx, -1, INT64_MIN, time_initial, INT64_MAX, 0) < 0) {
        fprintf(stderr, "%s: Unable to start from position %0.3f\n",
        vs->filename, (double)time_initial / AV_TIME_BASE);
     }
  }

  for(i=0; i<avf_ctx->nb_streams; i++)
     avf_ctx->streams[i]->discard = AVDISCARD_ALL;

  // Find best video stream
  if(disable_video==0) {
     video_index = av_find_best_stream(avf_ctx, AVMEDIA_TYPE_VIDEO,
                                       -1, 
                                       -1, 
                                       NULL, 0);
  }
  // Find best audio stream
  if(disable_audio==0) {
     audio_index = av_find_best_stream(avf_ctx, AVMEDIA_TYPE_AUDIO,
                                       -1,
                                       video_index,
                                       NULL, 0);
  }

/*
  for(i=0; i<avf_ctx->nb_streams; i++) {
    avf_ctx->streams[i]->discard = AVDISCARD_ALL;

    avctx = avf_ctx->streams[i]->codec;
  // Find the first audio stream
      if(avctx->codec_type == AVMEDIA_TYPE_AUDIO && audio_index < 0) {
        audio_index=i;
        fprintf(stderr, "found audio stream at index: %i\n", audio_index);
      }
  // Find the first video stream
      if(avctx->codec_type == AVMEDIA_TYPE_VIDEO && video_index < 0) {
        video_index=i;
        fprintf(stderr, "found video stream at index: %i\n", video_index);
      }
  }
*/

  // Open the audio stream
  if(audio_index >= 0 && disable_audio == 0) {
     fprintf(stderr, "found Audio stream at index %i\n", audio_index);
     stream_open(vs, audio_index);
  }
  // Open the video stream
  if(video_index >= 0 && disable_video == 0) {
     fprintf(stderr, "found Video stream at index %i\n", video_index);
     stream_open(vs, video_index);
  }  

  vs->refresh_tid = SDL_CreateThread(refresh_thread, vs);

  if(vs->video_stream < 0 && vs->audio_stream < 0) {
     fprintf(stderr, "%s: Unable to open codecs\n", vs->filename);
     goto fail;
  }

  // the main decoding loop
  for(;;) {
     if(vs->quit) {
        break;
     }
     if (vs->pause_curr != vs->pause_last) {
        vs->pause_last = vs->pause_curr;
        if(vs->pause_curr) {
           vs->pause_back = av_read_pause(avf_ctx);
        }
        else {
           av_read_play(avf_ctx);
        }
     } 

     // the seeking subroutine 
     if(vs->seek_req) {

        int64_t seek_target = vs->seek_pos;
        int64_t seek_min = vs->seek_rel > 0 ? seek_target - vs->seek_rel + 2: INT64_MIN;
        int64_t seek_max = vs->seek_rel < 0 ? seek_target - vs->seek_rel - 2: INT64_MAX;

        ret = avformat_seek_file(vs->avf_ctx, -1, seek_min, seek_target, seek_max, 0);
        if(ret < 0) {
	       fprintf(stderr, "%s: error while seeking\n", vs->avf_ctx->filename);
        } else {
		   // send audio packets to queue
	       if(disable_audio == 0 && vs->audio_stream >= 0) {
	          packet_queue_flush(&vs->audioq);
	          packet_queue_put(&vs->audioq, &flush_pkt);
              fprintf(stderr, "Seeking on audio stream: %i\n", vs->audio_stream);
	       }
		   // send video packets to queue
           if(disable_video == 0 && vs->video_stream >= 0) {
	          packet_queue_flush(&vs->videoq);
	          packet_queue_put(&vs->videoq, &flush_pkt);
              fprintf(stderr, "Seeking on video stream: %i\n", vs->video_stream);
	       } 
        }
        vs->seek_req = 0;
        eof= 0;
    }

    // if the queue are full, no need to read more
    if(vs->videoq.size + vs->audioq.size > MAX_TOTAL_QUEUE_SIZE ||
      ((vs->audioq.size > MAX_AUDIO_QUEUE_SIZE || vs->audio_stream < 0 ) &&
       (vs->videoq.nb_packets > MAX_VIDEO_QUEUE_FRAMES || vs->video_stream < 0 ))){
       // wait 10 msec
       SDL_Delay(10);
       continue;
    }

    // check if we are at the end and play in loop if the case
    if(eof) {
            if(disable_video == 0 && vs->video_stream >= 0){
                av_init_packet(packet);
                packet->data=NULL;
                packet->size=0;
                packet->stream_index= vs->video_stream;
                packet_queue_put(&vs->videoq, packet);
            } 
            if(disable_audio == 0 && vs->audio_stream >= 0 &&
               vs->audio_st->codec->codec->capabilities & CODEC_CAP_DELAY){
                av_init_packet(packet);
                packet->data=NULL;
                packet->size=0;
                packet->stream_index= vs->audio_stream;
                packet_queue_put(&vs->audioq, packet);
            } 
            // wait 10 msec
            SDL_Delay(10);

            if(vs->audioq.size + vs->videoq.size == 0){
                if(loop == 1 ){
                    stream_seek(vs, time_s != 0 ? time_s : 0, 0);
                }else if(loop == 0 ){
                    ret=AVERROR_EOF;
                    goto fail;
                }
            }
            eof=0;
            continue;
    }

    ret = av_read_frame(avf_ctx, packet);
    if(ret < 0) {
        if(url_feof(avf_ctx->pb) || ret == AVERROR_EOF) {
             eof=1;
        }    
        if(avf_ctx->pb && avf_ctx->pb->error) {
             break;
        }
        SDL_Delay(100); // wait 100 msec for user event
        continue;
    }

    // check if packet is in play range specified by user, then queue, otherwise discard 
    if (time_d != 0) {
    pkt_user_range = 
       (packet->pts - avf_ctx->streams[packet->stream_index]->start_time) *
       av_q2d(avf_ctx->streams[packet->stream_index]->time_base) -
       (double)(time_s != 0 ? time_s : 0) * 1e-6
       <= ((double)time_d * 1e-6);
    }
    // check if packet is from the audio or video stream
    if(packet->stream_index == vs->audio_stream && pkt_user_range != 0)
       packet_queue_put(&vs->audioq, packet);
    else if(packet->stream_index == vs->video_stream && pkt_user_range != 0)
       packet_queue_put(&vs->videoq, packet);
    else
       av_free_packet(packet);
  }

  // wait until the end
  while(!vs->quit) {
     SDL_Delay(100);
  }
  ret = 0;

 fail:
    // disable interrupting
    global_video_state = NULL;

    if (vs->audio_stream >= 0)
       stream_close(vs, vs->audio_stream);      
    if (vs->video_stream >= 0)
       stream_close(vs, vs->video_stream);
    if (vs->avf_ctx) {
       avformat_close_input(&vs->avf_ctx);
    } 
    //avio_set_interrupt_cb(NULL);
    avf_ctx->interrupt_callback.callback = NULL;

    if(ret != 0) {
       SDL_Event event;
       event.type = FF_QUIT_EVENT;
       event.user.data1 = vs;
       SDL_PushEvent(&event);
    }
    return 0;
}

static void do_video_frame_drop(VideoState *vs) {
    if (drop==0) {
      drop=1;
      fprintf(stderr, "video frame drop: enabled\n");
	} else if (drop==1) {
      drop=0;
      fprintf(stderr, "video frame drop: disabled\n");
    }
    video_refresh_timer(vs);
    //video_screen_open(vs);
}

static void do_full_screen(VideoState *vs) {
    if (full==0) 
      full=1;
	else if (full==1)
      full=0;
    video_screen_open(vs);
}

static void do_cycle_audio_ch(VideoState *vs) {
    int str_index;
    AVFormatContext *avf_ctx = vs->avf_ctx;
    AVCodecContext *dec = vs->audio_st->codec;
    AVStream *st = vs->audio_st;
 
    if (disable_audio==0 && st->codec->sample_rate != 0) {
      str_index=vs->audio_stream;
      if (cycle_audio_ch==0) {
        cycle_audio_ch=1;
        dec->channel_layout = 1;
        av_get_channel_layout_nb_channels(dec->channel_layout);
        fprintf(stderr, "audio channel: %i\n", av_get_channel_layout_nb_channels(dec->channel_layout));
	  } else if (cycle_audio_ch==1) {
        cycle_audio_ch=0;
        dec->channel_layout = 0;
        av_get_channel_layout_nb_channels(dec->channel_layout);
        fprintf(stderr, "audio channel: %i\n", av_get_channel_layout_nb_channels(dec->channel_layout));
      }
      stream_close(vs, str_index);
      stream_open(vs, str_index);
    }
}

static void do_enlarge_h_picture(VideoState *vs) {
    switch (enlarge_h) {
      case 0: 
        enlarge_h=1;
        break;
      case 1: 
        enlarge_h=2;
        break;
      case 2: 
        enlarge_h=3;
        break;
      case 3: 
        enlarge_h=4;
        break;
      case 4: 
        enlarge_h=5;
        break;
      case 5: 
        enlarge_h=6;
        break;
      case 6: 
        enlarge_h=0;
        break;
    }
    video_screen_open(vs);
}

static void do_enlarge_v_picture(VideoState *vs) {
    switch (enlarge_v) {
      case 0: 
        enlarge_v=1;
        break;
      case 1: 
        enlarge_v=2;
        break;
      case 2: 
        enlarge_v=3;
        break;
      case 3: 
        enlarge_v=4;
        break;
      case 4: 
        enlarge_v=5;
        break;
      case 5: 
        enlarge_v=6;
        break;
      case 6: 
        enlarge_v=0;
        break;
    }
    video_screen_open(vs);
}

// if the stream is paused unpause it, then step
static void step_to_next_frame(VideoState *vs) {

    if (vs->pause_curr)
       do_pause(vs);
    vs->step = 1;
}


static void do_sws_flags(VideoState *vs, int sflag) {
    int str_index;
           switch(sflag){
              case 0:
                 sws_flags = SWS_FAST_BILINEAR;
                 break;
              case 1:
                 sws_flags = SWS_BILINEAR;
                 break;
              case 2:
                 sws_flags = SWS_BICUBIC;
                 break;
              case 3:
                 sws_flags = SWS_X;
                 break;
              case 4:
                 sws_flags = SWS_POINT;
                 break;
              case 5:
                 sws_flags = SWS_AREA;
                 break;
              case 6:
                 sws_flags = SWS_BICUBLIN;
                 break;
              case 7:
                 sws_flags = SWS_GAUSS;
                 break;
              case 8:
                 sws_flags = SWS_SINC;
                 break;
              case 9:
                 sws_flags = SWS_LANCZOS;
                 break;
              case 10:
                 sws_flags = SWS_SPLINE;
                 break;
           }
           fprintf(stderr, "SWS: %i\n", sws_flags);

    if (disable_video==0) {
        str_index=vs->video_stream;
        stream_close(vs, str_index);
        stream_open(vs, str_index);
    }
}


// Set image filter correction values
static void do_img_correction(VideoState *vs, int fflag){
    int str_index;
    switch(fflag) {
    case IF_LUMA_GB:
        vs->luma_GBlur+=0.1;
        if(vs->luma_GBlur > 10) vs->luma_GBlur = 0;
        fprintf(stderr, "LumaGBlur: %f\n", vs->luma_GBlur);
        break;
    case IF_CHRO_GB:
        vs->chro_GBlur+=0.1;
        if(vs->chro_GBlur > 10) vs->chro_GBlur = 0;
        fprintf(stderr, "ChromaGBlur: %f\n", vs->chro_GBlur);
        break;
    case IF_LUMA_S:
        vs->luma_Sharpen+=0.1;
        if(vs->luma_Sharpen > 10) vs->luma_Sharpen = 0;
        fprintf(stderr, "LumaSharpen: %f\n", vs->luma_Sharpen);
        break;
    case IF_CHRO_S:
        vs->chro_Sharpen+=0.1;
        if(vs->chro_Sharpen > 10) vs->chro_Sharpen = 0;
        fprintf(stderr, "ChromaSharpen: %f\n", vs->chro_Sharpen);
        break;
    case IF_CHRO_HS:
        vs->chro_HShift+=0.1;
        if(vs->chro_HShift > 10) vs->chro_HShift = 0;
        fprintf(stderr, "ChromaHShift: %f\n", vs->chro_HShift);
        break;
    case IF_CHRO_VS:
        vs->chro_VShift+=0.1;
        if(vs->chro_VShift > 10) vs->chro_VShift = 0;
        fprintf(stderr, "ChromaVShift: %f\n", vs->chro_VShift);
        break;
    case IF_NOPTS:
        break;
    default:
        break;
    }

    vs->img_filter = sws_getDefaultFilter(vs->luma_GBlur,   vs->chro_GBlur,
                                          vs->luma_Sharpen, vs->chro_Sharpen,
                                          vs->chro_HShift,  vs->chro_VShift,
                                          0);

    if (disable_video==0) {
        str_index=vs->video_stream;
        stream_close(vs, str_index);
        stream_open(vs, str_index);
    }
}


/* loop to handle an event sent by the GUI */
static int event_loop(VideoState *video_state, int bpp) {
    SDL_Event event;
    double incr, pos;
    int sflag = 0;
    VideoState *vs = video_state;

    for(;;) {
    SDL_WaitEvent(&event);
    switch(event.type) {
    case SDL_KEYDOWN:
      switch(event.key.keysym.sym) {
         case SDLK_ESCAPE:
         case SDLK_q: // q: quit SDL loop
           do_exit(vs);
           return 1;
           break;
         case SDLK_SPACE:
         case SDLK_p: //p: pause playing
           do_pause(vs);
           break;
         case SDLK_s: //s: step to next frame
           step_to_next_frame(vs);
           break;
         case SDLK_c: //c: cicle audio channels
           do_cycle_audio_ch(vs);
           break;
         case SDLK_f: //f: fullscreen size
           do_full_screen(vs);
           break;
         case SDLK_d: //d: video frame drop
           do_video_frame_drop(vs);
           break;
         case SDLK_h: //h: enlarge horizontally the picture on the screen in case of wrong aspect ratio
           do_enlarge_h_picture(vs);
           break;
         case SDLK_v: //v: enlarge vertically the picture on the screen in case of wrong aspect ratio
           do_enlarge_v_picture(vs);
           break;
         case SDLK_o: // o: original picture size
           screen_w = vs->width = vs->video_st->codec->width;
           screen_h = vs->height = vs->video_st->codec->height;
           screen = SDL_SetVideoMode (screen_w, screen_h, bpp, SDL_HWSURFACE|SDL_ASYNCBLIT|SDL_HWACCEL);
           screen_w = screen_h = 0;
           enlarge_h = 0;
           enlarge_v = 0;
           imgcommand = ICON_SO;
           show_command = 1;
           break;
         case SDLK_g: // g: greater picture size
           if(!screen_w) screen_w = vs->video_st->codec->width;
           if(!screen_h) screen_h = vs->video_st->codec->height;
           screen_w = vs->width = screen_w * 1.1;
           screen_h = vs->height = screen_h * 1.1;
           screen = SDL_SetVideoMode (screen_w, screen_h, bpp, SDL_HWSURFACE|SDL_ASYNCBLIT|SDL_HWACCEL);
           if(screen_w > vs->video_st->codec->width) 
             imgcommand = ICON_SG;
           else if(screen_w < vs->video_st->codec->width) 
             imgcommand = ICON_SS;
           else if(screen_w == vs->video_st->codec->width) 
             imgcommand = ICON_SO;
           show_command = 1;
           break;
         case SDLK_b: // b: smaller picture size
           if(!screen_w) screen_w = vs->video_st->codec->width;
           if(!screen_h) screen_h = vs->video_st->codec->height;
           screen_w = vs->width =  screen_w / 1.1;
           screen_h = vs->height = screen_h / 1.1;
           screen = SDL_SetVideoMode (screen_w, screen_h, bpp, SDL_HWSURFACE|SDL_ASYNCBLIT|SDL_HWACCEL);
           if(screen_w < vs->video_st->codec->width) 
             imgcommand = ICON_SS;
           else if(screen_w > vs->video_st->codec->width) 
             imgcommand = ICON_SG;
           else if(screen_w == vs->video_st->codec->width) 
             imgcommand = ICON_SO;
           show_command = 1;
           break;
         case SDLK_1:
           do_img_correction(vs, IF_LUMA_GB);
           break;
         case SDLK_2:
           do_img_correction(vs, IF_CHRO_GB);
           break;
         case SDLK_3:
           do_img_correction(vs, IF_LUMA_S);
           break;
         case SDLK_4:
           do_img_correction(vs, IF_CHRO_S);
           break;
         case SDLK_5:
           do_img_correction(vs, IF_CHRO_HS);
           break;
         case SDLK_6:
           do_img_correction(vs, IF_CHRO_VS);
           break;
         case SDLK_0:
		   sflag++;
           if(sflag>10)
             sflag=0;
           do_sws_flags(vs, sflag);
           break;
         case SDLK_LEFT:
	       incr = -15.0;
           imgcommand = ICON_B;
           show_command = 1;
	       goto do_seek;
         case SDLK_RIGHT:
	       incr = 15.0;
           imgcommand = ICON_F;
           show_command = 1;
	       goto do_seek;
         case SDLK_UP:
	       incr = 60.0;
           imgcommand = ICON_FF;
           show_command = 1;
	       goto do_seek;
         case SDLK_DOWN:
	       incr = -60.0;
           imgcommand = ICON_BB;
           show_command = 1;
	       goto do_seek;
         do_seek:
	       if(vs) {
	         pos = get_master_clock(vs);
	         pos += incr;
             stream_seek(vs, (int64_t)(pos * AV_TIME_BASE), (int64_t)(incr * AV_TIME_BASE));
           }
	     break;
       default:
	     break;
      }
      break;
    case SDL_VIDEORESIZE:
      screen = SDL_SetVideoMode(event.resize.w, event.resize.h, bpp,
                                SDL_HWSURFACE|SDL_ASYNCBLIT|SDL_RESIZABLE|SDL_HWACCEL);
      screen_w = vs->width = event.resize.w;
      screen_h = vs->height = event.resize.h;
      break;
    case SDL_QUIT:
    case FF_QUIT_EVENT:
      do_exit(vs);
      return 1;
      break;
    case FF_ALLOC_EVENT:
      video_screen_open(event.user.data1);
      alloc_picture(event.user.data1);
      break;
    case FF_REFRESH_EVENT:
      video_refresh_timer(event.user.data1);
      vs->refresh = 0;  // reset the refresh flag to be ready for the next
      break;
    default:
      break;
    }
  }

}

static int lockmgr(void **mtx, enum AVLockOp op) {
   switch(op) {
      case AV_LOCK_CREATE:
          *mtx = SDL_CreateMutex();
          if(!*mtx)
              return 1;
          return 0;
      case AV_LOCK_OBTAIN:
          return !!SDL_LockMutex(*mtx);
      case AV_LOCK_RELEASE:
          return !!SDL_UnlockMutex(*mtx);
      case AV_LOCK_DESTROY:
          SDL_DestroyMutex(*mtx);
          return 0;
   }
   return 1;
}

// parse file and call the GUI loop 
//                (filename,   starttime,  duration,   width, height, loop, full,  drop,  mute,  blind, audiodelay)
int loadPlayerFile(char *name, int64_t ts, int64_t td, int w, int h, int l, int f, int d, int a, int v, float ta){
  printf("start=%d, duration=%d, width=%i, height=%i, loop=%i, full=%i, drop=%i, mute=%i, blind=%i, audelay=%f\n", ts, td, w, h, l, f, d, a, v, ta);
  Uint32 sdl_init_flags;
  Uint32 sdl_video_flags;
  const SDL_VideoInfo *sdl_vi;
  int sdl_bpp;
  VideoState  *vs;
  int sw, sh;
  time_s = time_d = 0;
  // starttime, duration, audiodelay
  if(ts > 0) time_s = 1000000 * ts;
  if(td > 0) time_d = 1000000 * td; 
  if(ta != 0) time_a = ta / 200;
  // loop, fullscreen, framedrop
  if(l == 0 || l == 1) loop = l;
  if(f == 0 || f == 1) full = f;
  if(d == 0 || d == 1) drop = d;
  // disable audio or video
  if(a == 0 || a == 1) disable_audio = a;
  if(v == 0 || v == 1) disable_video = v;
/*
  // image filter correction
  if(lb<= 1 && lb >= 0.1) lumaGBlur = lb;
  if(cb<= 1 && cb >= 0.1) chroGBlur = cb;
  if(ls<= 1 && ls >= 0.1) lumaSharpen = ls;
  if(cs<= 1 && cs >= 0.1) chroSharpen = cs;
  if(ch<= 1 && ch >= 0.1) chroHShift = ch;
  if(cv<= 1 && cv >= 0.1) chroVShift = cv;
*/
  // paranoic check for name
  if(!name || name[0] != '/') {
      fprintf(stderr, "No suitable file specified!\n");
      return 1;
  }
  // allocate filename memory
  if (vs = av_mallocz(sizeof(VideoState))){ 
      stpncpy(vs->filename, name, sizeof(vs->filename));
  }
  // check for global structure
  if (!vs) {
      fprintf(stderr, "Unable to initialize struct VideoState!\n");
      do_exit(NULL);
  }

  // Register all codecs
#ifdef HAVE_LIBAVCODEC
  avcodec_register_all();
#endif
  // Register all codecs, demux and protocols
  av_register_all();
  // Initialize the network protocol
  avformat_network_init();

  // set SDL init flags
  sdl_init_flags |= SDL_INIT_VIDEO;
  if(disable_audio == 0)
      sdl_init_flags |= SDL_INIT_AUDIO;

  sdl_init_flags |= SDL_INIT_TIMER;
#if !defined(__MINGW32__) && !defined(__APPLE__)
  sdl_init_flags |= SDL_INIT_EVENTTHREAD;
#endif

  if(SDL_Init(sdl_init_flags)) {
      fprintf(stderr, "Unable to initialize SDL: %s\n", SDL_GetError());
      return 1;
  }

  // passed user video size
  if(w>=480 && h>=360) {
      sw = screen_w = w;
      sh = screen_h = h;
  }
  // recall previous video size if no values are passed on
  else if(screen_w>64 && screen_h>64) {
      sw = screen_w;
      sh = screen_h;
  }

  // find SDL full screen size
  if (disable_video == 0) {
#if HAVE_STRUCT_SDL_VIDEOINFOP
      sdl_vi = SDL_GetVideoInfo();
      fs_screen_w = sdl_vi->current_w;
      fs_screen_w = sdl_vi->current_h;
#endif
  }
  // find SDL bpp mode
#if HAVE_STRUCT_SDL_VIDEOINFOP
      sdl_vi = SDL_GetVideoInfo();
      switch (sdl_vi->vfmt->BitsPerPixel) {
          case 16:
          case 24:
          case 32:
              sdl_bpp = sdl_vi->vfmt->BitsPerPixel;
              break;
          default:
              sdl_bpp = 16;
              break;
      }
#else
      sdl_bpp = 0;
#endif
	
  // set SDL video flags
  sdl_video_flags = SDL_HWSURFACE|SDL_ASYNCBLIT|SDL_HWACCEL;

  if(full==1) {
    sdl_video_flags |= SDL_FULLSCREEN;
    if (full==1 && fs_screen_w) {
      sw = fs_screen_w;
      sh = fs_screen_h;
    }
  } else if(full==0) {
    sdl_video_flags |= SDL_RESIZABLE;
  }

  // set SDL video mode
  screen = SDL_SetVideoMode(sw, sh, sdl_bpp, sdl_video_flags);
  if (screen == NULL) {
      fprintf(stderr, "Unable to set w=%i, h=%i, bpp=%i SDL video mode: %s\n", sw, sh, sdl_bpp, SDL_GetError());
      return 1;
  }
  else {
      fprintf(stderr, "width=%i, height=%i, bpp=%i SDL video mode\n", sw, sh, sdl_bpp);
  }
  // reset
  sw = sh = 0;

  SDL_EventState(SDL_ACTIVEEVENT, SDL_IGNORE);
  SDL_EventState(SDL_SYSWMEVENT, SDL_IGNORE);
  SDL_EventState(SDL_USEREVENT, SDL_IGNORE);

  if (av_lockmgr_register(lockmgr)) {
     fprintf(stderr, "Unable to initialize lock manager!\n");
     do_exit(NULL);
  }

  av_init_packet(&flush_pkt);
  flush_pkt.data = "FLUSH";

  // get image filter correction values
  vs->luma_GBlur = lumaGBlur;
  vs->chro_GBlur = chroGBlur;
  vs->luma_Sharpen = lumaSharpen;
  vs->chro_Sharpen = chroSharpen;
  vs->chro_HShift = chroHShift;
  vs->chro_VShift = chroVShift;

  // Sincronize with audio clock first, then try video clock
  if (disable_audio == 0)
    av_sync_type = AV_SYNC_AUDIO_MASTER;
  else if (disable_video == 0)
    av_sync_type = AV_SYNC_VIDEO_MASTER;
  else 
    av_sync_type = AV_SYNC_EXTERNAL_MASTER;

  // Create the mutex
  vs->pictq_mutex = SDL_CreateMutex();
  // Create the condition
  vs->pictq_cond = SDL_CreateCond();

  vs->av_sync_type = av_sync_type;

  // Create and run the thread
  vs->decode_tid = SDL_CreateThread(decode_thread, vs);
  if(!vs->decode_tid) {
     av_free(vs);
     return -1;
  }
  // Set image filter correction values
  sws_freeFilter(vs->img_filter);
  //do_img_correction(vs, IF_NOPTS);

  event_loop(vs,sdl_bpp);

  // never return
  return 0;
}
